\ CONTENT     : Handles the main program of the factoring,
\               around the SolutionBase
\ LANGUAGE    : tForth
\ PROJECT     : hcc primes
\ CATEGORY    : EXAMPLE
\ AUTHOR      : AH 911108
\ LAST CHANGE : AH 931215 Better stop method
\ LAST CHANGE : AH 931129 Array of channel addresses, ALT input
\ LAST CHANGE : AH 931121 Cleanup protocol, mill only turns right
\ LAST CHANGE : MH 911121 Made it use 3 links and a memory process.
\ LAST CHANGE : MH 911120 SendReset using SELECT
\ LAST CHANGE : AH 911119 CalculateMaxBase removed
\ LAST CHANGE : 911119 MHX hardware link version
\ LAST CHANGE : AH 911119 Beep for a trivial solution, return to TEXT
\ LAST CHANGE : AH 911119 FindSolution used, stopper added
\ LAST CHANGE : AH 911118 added the pseudo parallelism
\ LAST CHANGE : AH 911112 added the solutions database
\ LAST CHANGE : AH 911108 created
\ NOTE : A dirty trick is being done with AT-XY to suppress all output.
\ In order for this to work
\   - all output is to be done in statements after AT-XY on the same line
\   - the remainder of the line after AT-XY should leave the stack
\      at the same depth as before

        NEEDS arrays
        REVISION -main "--- HCC main program    Version 4    ---"

        PRIVATES


%sizeof SOL1         CONSTANT #SOL PRIVATE
MaxB BITS/CELL * 1+  CONSTANT LenSolBase PRIVATE

10 CONSTANT MaxChannels
VALUE ChannelsUsed                   CLEAR ChannelsUsed

\ Address of input channel
MaxChannels ARRAY MainIn[]

\ Address of output channel
MaxChannels ARRAY MainOut[]

\ The process hanging off the channel waits for work
MaxChannels ARRAY WaitWork[]

\ The colors associated with each transputer
MaxChannels ARRAY TpColor[]

CreateFactorization TheFactorization


\ FetchSolution fetches the solution with index I (0..LenSolBase)
\ to address A, address of solution base is S

: FetchSolution ( A,I,S -- )
    SWAP #SOL * +       \ Source
    SWAP #SOL CMOVE
; PRIVATE


\ StoreSolution stores the solution with address A
\ to index I (0..LenSolBase) in the SolutionBase
\ address of solution base is S

: StoreSolution ( A,I,S -- )
    SWAP #SOL * +       \ Destination
    #SOL CMOVE
; PRIVATE


\ ?EmptySolution returns whether the entry with index I (0..LenSolBase)
\ has already been filled in, address of solution base is S

: ?EmptySolution ( I,S -- )
    SWAP #SOL * +       \ Entry address
    MaxB CELLS + @ 0=   \ V1 = 0 : This entry was not filled in
; PRIVATE


: CreateSolutionBase
    CREATE LenSolBase #SOL * DUP
    ALLOCATE ABORT" Not Enough memory for Solution Base"
    DUP ,
    SWAP ERASE
    DOES>
        @
        %VAR @  0 %VAR !     \ Reset to FROM
        CASE
          %clear%      OF LenSolBase #SOL * ERASE ENDOF
          %from%       OF FetchSolution           ENDOF
          %to%         OF StoreSolution           ENDOF
          %addressof%  OF ( Leave address)        ENDOF
          %sizeof%     OF DROP LenSolBase #SOL *  ENDOF
          %empty%      OF ?EmptySolution          ENDOF
          DUP DEC. TRUE ABORT" Solution: unknown message"
        ENDCASE
; PRIVATE


\ In the solution base the solutions I has a ms bit I.
\ A solution with ms bit 0 is also possible

CreateSolutionBase SolutionBase

\ PrintSolutionBase prints the solution base

: PrintSolutionBase ( -- )
    LenSolBase 0 DO
       I 0 =   0 %empty SolutionBase AND IF
         CR ." ENTRY 0 IS EMPTY"
       ENDIF
       I %empty SolutionBase 0= IF
          %addressof SolutionBase #SOL I * +
          PrintSolution ."  \ " I .  EOL
       ENDIF
   LOOP
;

\  Print, making sure previous output at the same place is erased.
\  (Erase to end of line AND the complete next line)
\ This subroutine is obsolete ( EOL CR EOL ) does the job
: MY-EOL        \ <> -- <>
                C/L ?AT >S - ( INVERS) SPACES
                0 S> 1+ AT-XY  C/L SPACES ( -INVERS)
; PRIVATE

0 [IF]


\ GPrintSolutionBase prints the solution base graphically
\ This subroutine is obsolete
: GPrintSolutionBase ( -- )
        LenSolBase 0 ?DO
             I %empty SolutionBase
             0= IF
                   %addressof SolutionBase #SOL I * +
                   MaxB BITS/CELL *
                   I Yoffs +
                   LINEDRAW-BITMAP
             ENDIF
         LOOP ;
[THEN]

\ Puts the solution SOL1 in the database.
\ If its place is occupied, it may ripple through to the top.

: PutSolutionInDB ( -- )
    %addressof SOL1 LOCAL S1
    %addressof SOL2 LOCAL S2
    %addressof SOL3 LOCAL S3
    BEGIN
            S1 FindMsBit
            [ USE-GRAPHICS ] [IF] DUP GPutArrow [THEN]
        %empty SolutionBase 0= WHILE \ Place where S1 belongs is occupied
            S2 S1 FindMsBit %from SolutionBase \ Get that solution to S2
            CombineSolution     \ Combine S1 and S2 into S3
            S3 S1 SSmove        \ Replace, S1 has now fewer bits
    REPEAT

    S1 ?NonTrivial IF
        S1 S1 FindMsBit [ USE-GRAPHICS ] [IF] DUP GShowSolution [THEN]
        %to SolutionBase
        0 2 AT-XY S1 PrintSolution ."  \ " S1 FindMsBit . MY-EOL
\     ELSE
\         7 EMIT          \ Denounce the false prophet
    ENDIF

    [ USE-GRAPHICS 0= DEBUG AND ] [IF]  PrintSolutionBase [THEN]

; PRIVATE


VALUE mill      PRIVATE         CLEAR mill

: .MILL     \ <> -- <>
[ USE-GRAPHICS ] [IF]
        ?AT
           mill 3 AND CASE
              0 OF &| ENDOF
              1 OF &/ ENDOF
              2 OF &- ENDOF
              3 OF &\ ENDOF
           ENDCASE
           C/L 8 -  L/SCR 1-  AT-XY DUP EMIT    \ See remark up front!
           DROP
           1 +TO mill
        AT-XY
[THEN]
; PRIVATE



\ FindSolution demonstrates how this mechanism could work
\ without sieving, by just trying everything.
\ The sieving highly speeds this up.
\ Here it is used to make sure the lowest entries are used.
\ A sharp criterion could make them fall through the sieve.
: FindSolution ( -- flag )
    %addressof SOL1 LOCAL S1
    %addressof SOL2 LOCAL S2
    %addressof SOL3 LOCAL S3

[ USE-GRAPHICS ] [IF]
    yellow TO bgcolor
[THEN]

    FALSE
    10 1 DO
        I S>D MakeSolution IF
           PutSolutionInDB
           \ Is the entry 0 filled in? That is almost a factor!
           0 %empty SolutionBase 0= IF
              DROP TRUE LEAVE
           ENDIF
        ENDIF
    LOOP ;




\ CalcCrit returns the amount of bits in factors before we decide it is a
\ solution candidate.
\ For the moment we want a part as large as TheRoot falling apart in small
\ factors. ( #sharp = 1000)

: CalcCrit   ( -- )
     TheRoot Vcount-bits #sharp 1000 */
;

DVALUE NextOffset       CLEAR NextOffset        \ Next offset to deal out
VALUE  Proposed

\ Deals out an amount of work to the process hanging off the
\  channel pair ("link") with number n
: GiveWork      ( <n> -- )
        DUP MainOut[] LOCAL ChanOut
            MainIn[]  LOCAL ChanIn

        %>offset   ChanOut CHANNEL-!
        NextOffset ChanOut CHANNEL-2!
        \ Double the length of the sieve but within bounds
        Proposed 2* DUP 0> IF TO Proposed ELSE DROP THEN
        Proposed ChanOut CHANNEL-!
        ChanIn CHANNEL-@ S>D +TO NextOffset
        100 0 AT-XY ." Offset:" NextOffset D.
;

\ Receive a solution over link `n'
: ReceiveSolution               ( <n> -- )
    LOCAL N
\   ." Waiting in MainReceive"
    N MainIn[] CHANNEL-@ CASE
\      CR ." Received " DUP . CR
       %<work     OF N GiveWork ENDOF
       %<solution OF
                     N MainIn[] CHANNEL-2@
                     MakeSolution IF PutSolutionInDB ELSE .MILL ENDIF
                  ENDOF
       DUP DEC. TRUE ABORT" ReceiveSolution: unknown message"
    ENDCASE
\   {{ CR ." MAIN" .S }}
; PRIVATE

: StartSievers                 ( -- )
        CalcCrit LOCAL Crit

        ChannelsUsed 0 DO
            %>crit     I MainOut[] CHANNEL-!
            Crit       I MainOut[] CHANNEL-!

            %>pbase    I MainOut[] CHANNEL-!
            %addressof PrimeBase  MaxPrime 3 * 2+ CELLS
                       I MainOut[] CHANNEL-SEND

            I GiveWork
\          {{ CR ." MAIN" .S }}
        LOOP
; PRIVATE

\ Throw away all solution messages and the following work message
: PurgeChannels   \ <> -- <>
        ChannelsUsed 0 DO
           BEGIN
              I MainIn[] Channel-@ DUP %<solution =
           WHILE
              DROP
              I MainIn[] CHANNEL-2@ 2DROP
           REPEAT
           %<work <> IF ABORT" work expected" THEN
        LOOP
;

\ Aborts if the user types a 'Q'
: ?ABORT        ( -- )
        KEY? IF
           KEY &Q = IF
              PurgeChannels
              TRUE ABORT" aborted by user interrupt"
           ELSE
              7 EMIT
           THEN
        THEN
;

: ReceiveSolutions       \ read from parallel processes
        SELECT
          ChannelsUsed 0 REPLICATE I MainIn[] GUARD
[ USE-GRAPHICS ] [IF]
              DUP TpColor[] TO bgcolor
[THEN]
              ReceiveSolution
           ENDGUARD
           16000 TIMEOUT \ Don't be dead for more than about one second
           ENDGUARD
        ENDSELECT
        ?ABORT
;


\ The testing whether TheNumber has a pure root is necessary
\ because the quadatric sieving cannot handle those numbers.
\ Composite powers (e.g. 6) will be found in two steps.
\ If the #SmallPrimesLimit is 10^5 no problems will happen
\ before 85 decimal digits.
\ Shooting SS towards the factorization does no harm
\  if SS is not a factor.
: FactorizeByPureRoots                    ( -- )
    0 1 AT-XY ." Partial: " %print TheFactorization EOL
    0 2 AT-XY ." Factoring Pure Roots: " TheNumber .VNUM
    TheNumber  2 VRoot    SS %factor TheFactorization
    TheNumber  3 VRoot    SS %factor TheFactorization
    TheNumber  5 VRoot    SS %factor TheFactorization
    TheNumber  7 VRoot    SS %factor TheFactorization
    TheNumber 11 VRoot    SS %factor TheFactorization
    TheNumber 13 VRoot    SS %factor TheFactorization
;

\ FactorizeBySieve factorises the number in TheNumber .
\ It must be composite, and it must not be a pure power, i.e.
\ FactorizeByPureRoots must have failed.
: FactorizeBySieve  ( -- )

    0.0 TO NextOffset
    1024 TO Proposed    \ Initial sieve length
    CLS
    0 0 AT-XY TheNumber .VNUM
    0 1 AT-XY ." Partial: " %print TheFactorization EOL
    0 2 AT-XY ." Creating Prime Base for: " TheNumber .VNUM
    TheNumber VSqrt  SS TheRoot VVmove  \ Remember it
    CreatePrimeBase
    0 2 AT-XY ." Factoring by Quadratic Sieving: " TheNumber .VNUM
    %clear SolutionBase

    FindSolution DROP

    StartSievers

    BEGIN       \ Is the entry 0 filled in? That is almost a factor!
      ReceiveSolutions
      0 %empty SolutionBase  0=
    UNTIL

    PurgeChannels

    %addressof SOL1 0 %from SolutionBase
    %V1 SOL1 %V2 SOL1 VV+               \ TheNumber = (V1-V2)(V1+V2) * N
    %V1 SOL1 TheNumber VGCD
    SS 1 VS= ABORT" A non-trivial 0 solution gives GCD=1 "
    SS %factor TheFactorization
;

: FACTORIZE
    [ USE-GRAPHICS ] [IF]
        INIT-GRAPHICS  CLS
    [THEN]
     TheNumber VREAD
     0 0 AT-XY TheNumber .VNUM
     TheNumber %to TheFactorization
     2 %prime TheFactorization
     #SmallPrimesLimit 3 DO
         I ?Prime IF
            I %prime TheFactorization
            %?factorized TheFactorization IF LEAVE THEN
         THEN
     2 +LOOP

     BEGIN
          %?factorized TheFactorization 0= WHILE
          TheNumber %todo TheFactorization
          FactorizeByPureRoots
          %?factorized TheFactorization 0= IF
              TheNumber %todo TheFactorization
              FactorizeBySieve
          THEN
     REPEAT

     7 EMIT  0 4 AT-XY ." Total: " %print TheFactorization EOL

;
                DEPRIVE

                                (* End of File *)
