\ COPYRIGHT 1993 DUTCH FORTH WORKSHOP - FILENAME : tserver.frt #1
\ LANGUAGE    : ANS Forth with environmental dependancies
\ PROJECT     : tForth
\ DESCRIPTION : t-Server
\ CATEGORY    : Utilities
\ AUTHOR      : Erwin Dondorp
\ LAST CHANGE : 930521 AH the return code of the STOP command was not fetched
\ LAST CHANGE : 930521 AH cosmetic
\ LAST CHANGE : 930502 AH use of BOOTLINK-IN BOOTLINK-OUT
\ LAST CHANGE : May 1st, 1993  Marcel Hendrix, use-in use-out etcetera
\ LAST CHANGE : April 16, 1993  Marcel Hendrix
\ LAST CHANGE : June 24th, 1992  Marcel Hendrix
\


DOC
(*
 TODO:
 Do not reset a transputer if it is booting (already reset) since the reset on
 some hardware can reset other transputers as well.

 This is the SERVER implemented as a bridge.
 The server cannot be implemented as a repeater since there is no uniform
 packet format.
 The server merely copies the commands to its bootlink and returns the result
 on the same link as the request was on.
 Only a few commands are intercepted and acted on specially:

 TERM 0  : to enable our own command parsing, but usually send through;
 TERM 17 : to enable our own command parsing, but usually send through;
 EXE 15  : send 'T' as server ID, never passed;
 FORTH   : executed by this server;
 DATA    : Passed to a vectored routine;
 BYE     : leaving this server only, never passed.

 CLIENT is DOWN
 SERVER is UP

 The main routine starts a new tForth on one of its links.

 This code has worked with two Parsytec TPM-PC boards, connected via
 LINK2 with internal release version 1.1.E.
 It could not be tested with the official release 1.1 (for lack
 of hardware), but it should work okay.
 To be sure, check the problem notes of release 1.1.

*)
ENDDOC


        REVISION -tserver " Transputer server   Version 3.03t"

        PRIVATES


VALUE networktype	0 TO networktype	-- 0 unknown network
                                                -- 1 CSA
                                                -- 2 Parsytec

VALUE theend			PRIVATE
VALUE ProcessorID		PRIVATE
VALUE bootfileid		PRIVATE
VALUE use-in			PRIVATE
VALUE use-out			PRIVATE
VALUE filename			PRIVATE 	0 TO filename
VALUE filenamelen		PRIVATE 	0 TO filenamelen
VALUE MemStart			PRIVATE
VALUE binfile			PRIVATE
VALUE boot			PRIVATE 	2 TO boot
VALUE growsize			PRIVATE 	0 TO growsize
VALUE memsize			PRIVATE 	0 TO memsize
CREATE cmdl 	   #1000 ALLOT	PRIVATE
VALUE cmdlptr			PRIVATE 	cmdl TO cmdlptr
VALUE cmdp			PRIVATE 	cmdl TO cmdp
VALUE resetvector		PRIVATE

VALUE datahandler		PRIVATE		0 TO datahandler

CREATE tbuffer	  #10000 ALLOT	PRIVATE
CREATE bootfiletext #256 ALLOT	PRIVATE
CREATE cmdinc	   #1000 ALLOT	PRIVATE
VALUE cmdincptr			PRIVATE 	cmdinc TO cmdincptr
CREATE cmdextra	   #1000 ALLOT	PRIVATE
VALUE cmdextraptr		PRIVATE 	cmdextra TO cmdextraptr

CREATE codeinfo  5 CELLS ALLOT	PRIVATE
CREATE XYZ	 5 CELLS ALLOT	PRIVATE

: codeinfo>	EVAL" codeinfo []CELL @ " ; IMMEDIATE PRIVATE


\ Rely on proper path / append

CREATE tfboot$	," tfboot.x"	PRIVATE
CREATE t200$    ," tf2_____.x"	PRIVATE
CREATE t400$    ," tf4fde__.x"	PRIVATE
CREATE t800$    ," tf8fd___.x"	PRIVATE

CREATE tids	\ table indexed by TID (0...9)
                0    ,	    0 ,	  #200 ,   0 ,
                #414 ,	 #425 ,	     0 ,   0 ,
                #800 ,	 #805 ,
        PRIVATE


WARNING @  WARNING OFF

 : _RX	EVAL" BOOTLINK-IN  CHANNEL-C@ " ;      IMMEDIATE PRIVATE
 : _TX	EVAL" BOOTLINK-OUT CHANNEL-C! " ;      IMMEDIATE PRIVATE
 : _RXW EVAL" BOOTLINK-IN  CHANNEL-@ "  ;      IMMEDIATE PRIVATE
 : _TXW EVAL" BOOTLINK-OUT CHANNEL-! "  ;      IMMEDIATE PRIVATE
 : _TX$ EVAL" BOOTLINK-OUT CHANNEL-SEND "  ;   IMMEDIATE PRIVATE
 : _RX$ EVAL" BOOTLINK-IN  CHANNEL-RECEIVE " ; IMMEDIATE PRIVATE


WARNING !


: n->UP		use-in CHANNEL-@  _TXW	; PRIVATE
: n->DOWN	_RXW use-out  CHANNEL-!	; PRIVATE

: $->UP		use-in CHANNEL-@ DUP >S
                ( count) _TXW
                tbuffer S use-in CHANNEL-RECEIVE
                tbuffer S> _TX$ ; PRIVATE

: $->DOWN	use-in CHANNEL-@   LOCAL max-size
                max-size _TXW  _RXW LOCAL act-size
                tbuffer act-size _RX$
                act-size use-out CHANNEL-!
                tbuffer act-size use-out CHANNEL-SEND ; PRIVATE


: QUIRKY-DOWN	use-in CHANNEL-@ DUP  _TXW
                tbuffer OVER _RX$
                tbuffer SWAP use-out CHANNEL-SEND ; PRIVATE


\ Doesn't stop, just terminates the serving action of the UP transputer.

: DO-STOP	TRUE TO theend ; PRIVATE
: IEXE		$FF _TX	  1 _TX   _TX ; PRIVATE		\ <cmd> --- <>
: ITERM		$FF _TX	  2 _TX   _TX ; PRIVATE		\ <cmd> --- <>
: IUSER		$FF _TX	  3 _TX   _TX ; PRIVATE		\ <cmd> --- <>


\ file and i/o commands

: DO-EXE
        use-in CHANNEL-C@ DUP
        CASE
          0 OF IEXE  $->UP n->UP n->DOWN n->DOWN ENDOF	( open)
 	  1 OF IEXE  $->UP n->DOWN		 ENDOF	( os <command>)
          2 OF IEXE  n->DOWN			 ENDOF	( os)
          3 OF IEXE  $->UP n->DOWN 		 ENDOF	( delete file)
          4 OF IEXE  n->UP n->UP n->UP n->DOWN 	 ENDOF	( resize)
          5 OF IEXE  n->UP n->DOWN 		 ENDOF	( flush)
          6 OF IEXE  $->UP n->UP n->DOWN n->DOWN ENDOF	( create)
          7 OF IEXE  n->UP n->DOWN 		 ENDOF	( close)
          8 OF IEXE  n->UP $->DOWN
                     n->DOWN n->DOWN 	 	 ENDOF	( read-line)
          9 OF IEXE  n->DOWN n->DOWN n->DOWN 	 ENDOF	( time@)
        #10 OF IEXE  n->DOWN n->DOWN n->DOWN 	 ENDOF	( date@)
        #11 OF IEXE  $->UP $->UP n->DOWN	 ENDOF	( rename)
        #12 OF IEXE  n->UP n->UP n->UP n->UP
                     n->DOWN n->DOWN n->DOWN 	 ENDOF	( seek)
        #13 OF IEXE  n->UP $->DOWN n->DOWN 	 ENDOF	( read-file)
        #14 OF IEXE  n->UP $->UP   n->DOWN 	 ENDOF	( write-file)
        #15 OF	     DROP 'T' use-out CHANNEL-!	 ENDOF	( Get server version)
        #16 OF IEXE  n->UP $->UP   n->DOWN 	 ENDOF	( write-line)
        #17 OF      DROP My# 3 * use-in $0F AND
                    2 RSHIFT + use-out CHANNEL-!
                    #Xptrs  1+ use-out CHANNEL-! ENDOF	( My# #Xptrs)
        #18 OF IEXE  6 0 DO n->DOWN LOOP	 ENDOF	( time&date)
        #19 OF IEXE  $->UP n->DOWN 		 ENDOF	( chdir)
        #20 OF IEXE  $->DOWN n->DOWN 		 ENDOF	( getdir)
                     CR DEC. TRUE ABORT" T-Server: Illegal exe-cmd"
        ENDCASE	; PRIVATE

: DO-TERM
        use-in CHANNEL-C@ DUP
        CASE
          0 OF
                cmdp cmdlptr <> IF
                        DROP
                        cmdp C@ use-out CHANNEL-!
                        1 +TO cmdp
                ELSE	ITERM n->DOWN
                ENDIF				ENDOF	( ekey)
          1 OF ITERM  n->UP n->UP 		ENDOF	( at-xy)
          2 OF ITERM  n->DOWN n->DOWN 		ENDOF	( ?at)
          3 OF ITERM  n->UP n->UP n->UP 	ENDOF	( !attribs)
          4 OF ITERM  n->DOWN 			ENDOF	( c/l)
          5 OF ITERM  n->DOWN 			ENDOF	( l/scr)
          6 OF ITERM  n->DOWN 			ENDOF	( c/scr)

          7 OF ITERM 				ENDOF	( 7..12 obsolete)
          8 OF ITERM 				ENDOF
          9 OF ITERM 				ENDOF
        #10 OF ITERM 				ENDOF
        #11 OF ITERM 				ENDOF
        #12 OF ITERM 				ENDOF

        #13 OF ITERM 				ENDOF	( cls)
        #14 OF ITERM				ENDOF	( eol)

        #15 OF ITERM 				ENDOF	( 15, 16 obsolete)
        #16 OF ITERM 				ENDOF

        #17 OF
                cmdp cmdlptr <> IF
                        DROP
                        TRUE use-out CHANNEL-!
                ELSE	ITERM n->DOWN
                ENDIF				ENDOF	( ekey?)
        #18 OF ITERM  n->DOWN 			ENDOF	( textmode)

        #19 OF ITERM 				ENDOF	( 19, 20 obsolete)
        #20 OF ITERM 				ENDOF

        #21 OF ITERM  $->UP 			ENDOF	( type)
        #22 OF ITERM 				ENDOF	( log-open)
        #23 OF ITERM 				ENDOF	( log-close)
        #26 OF ITERM  n->UP n->UP n->UP n->DOWN	ENDOF	( calc-rgb)
        #27 OF ITERM  n->UP 			ENDOF	( ms)
        #28 OF ITERM  n->DOWN 			ENDOF	( ?ms)
        #29 OF ITERM  n->UP 			ENDOF	( set cursor)
                      CR DEC. TRUE ABORT" T-Server: Illegal term-command"
        ENDCASE	; PRIVATE

: DO-USER
        use-in CHANNEL-C@ DUP
        CASE
         0 OF IUSER n->UP n->UP $->UP 		  ENDOF	( plot line)
         1 OF IUSER n->DOWN n->DOWN n->DOWN 	  ENDOF	( graphics)

         2 OF IUSER 				  ENDOF	( 2..4 obsolete)
         3 OF IUSER 			 	  ENDOF
         4 OF IUSER 				  ENDOF

         5 OF IUSER 				  ENDOF	( text)
         6 OF IUSER n->UP n->UP n->UP 		  ENDOF	( set-dot)
         7 OF IUSER n->UP n->UP n->UP n->UP n->UP ENDOF	( draw-line)
         8 OF IUSER n->UP 			  ENDOF	( select-hires)
         9 OF IUSER n->UP 			  ENDOF	( set-pen)
       #10 OF IUSER $->UP 			  ENDOF	( set-palette)
       #11 OF IUSER QUIRKY-DOWN 		  ENDOF	( get-palette)
       #12 OF IUSER $->UP 			  ENDOF	( set-font)
                    CR DEC. TRUE ABORT" T-Server: Illegal user-command"
        ENDCASE	; PRIVATE


: DO-FORTH	use-in CHANNEL-@ LOCAL data-len	\ do not pass
                tbuffer data-len use-in CHANNEL-RECEIVE
                tbuffer data-len EVALUATE ; PRIVATE


: DO-DATA	use-in CHANNEL-@ LOCAL data-len
                tbuffer data-len use-in CHANNEL-RECEIVE
                datahandler
        CASE
         0 OF ( ignore) 			ENDOF
        -1 OF ( send through) $FF _TX  5 _TX
                tbuffer data-len _TX$		ENDOF
                tbuffer data-len use-in datahandler EXECUTE
        ENDCASE	; PRIVATE


: DO-FF		$FF _TX	 $FF _TX ; PRIVATE


: SERVE-CMD

        SELECT
                use-in GUARD ENDGUARD
                #3000 >TICKS TIMEOUT TRUE ABORT" T-Server: timeout" ENDGUARD
        ENDSELECT
      {{
        use-in CHANNEL-C@ DUP $FF <> IF _TX }} EXIT ENDIF DROP
        use-in CHANNEL-C@
        CASE
          0 OF DO-STOP	ENDOF
          1 OF DO-EXE	ENDOF
          2 OF DO-TERM	ENDOF
          3 OF DO-USER	ENDOF
          4 OF DO-FORTH ENDOF
          5 OF DO-DATA	ENDOF
        $FF OF DO-FF	ENDOF
                }} TRUE ABORT" T-Server: Unknown server category"
        ENDCASE	}} ; PRIVATE


: SERVE	FALSE TO theend
        BEGIN SERVE-CMD  theend UNTIL
        use-in CHANNEL-C@       \ Fetch exit code
;


\ RESET -----------------------------------------

        0 =: 'Reset-c	PRIVATE -- Reset the [down linked] transputers here
        4 =: 'Analyze-c PRIVATE -- Analyze command goes here
        0 =: 'Error-c	PRIVATE -- Read error flag here.

$000000C0 =: 'Reset-p	PRIVATE -- for Parsytec TPM-PC

: FER'SURE?
        ." The t-Server needs to reset the requested computer, but while" CR
        ." doing so, more transputers than the one requested might be" CR
        ." reset. Are you sure you want to allow this? " KEY DUP EMIT >UPC CR
        'Y' <> ABORT" T-Server, User abort in Reset" ; PRIVATE

: RESET-LINK-C
        FER'SURE?
        0 'Reset-c C!	#50 MS		\ make sure reset is off,
        1 'Reset-c C!	#50 MS		\ assert for 50 ms,
        0 'Reset-c C!			\ deassert forever.
        ; PRIVATE

: ANALYZE-LINK-C
        FER'SURE?
        0 'Reset-c   C!
        0 'Analyze-c C!	  #50 MS
        1 'Analyze-c C!	 #150 MS
        1 'Reset-c   C!	  #50 MS
        0 'Reset-c   C!	 #150 MS
        0 'Analyze-c C!
        ; PRIVATE

\ Reset a Parsytec board on any of link 0, 1, 2, or 3.

: RESET-LINK-P ( -- )
        4 0 DO I 'Reset-p ! LOOP
        use-in
        CASE
         LINK0-IN OF 1 ENDOF
         LINK1-IN OF 2 ENDOF
         LINK2-IN OF 4 ENDOF
         LINK3-IN OF 8 ENDOF
        ENDCASE
        'Reset-p !
        1 MS		  \ recommended: 2 ticks = 128 usec.
        0 'Reset-p ! ; PRIVATE


\ DETERMINE ---------------------------------------------------

: bootfile
        tfboot$ COUNT R/O BIN OPEN-FILE SWAP TO bootfileid
                  ABORT" T-Server: can't find or open tfboot.x"
        bootfiletext #256 bootfileid READ-FILE
                  ABORT" T-Server: can't read tfboot.x"
        DUP  0=   ABORT" tfboot.x cannot be 0 bytes"
        DUP  1  = ABORT" tfboot.x cannot be 1 byte"
        DUP $FF > ABORT" tfboot.x cannot be larger than 255 bytes"
        {{
           DUP use-out CHANNEL-C!
           bootfiletext SWAP use-out CHANNEL-SEND
        }}
        bootfileid CLOSE-FILE ABORT" cannot close file tfboot.x" ; PRIVATE

: determine
        SELECT
          use-in GUARD ENDGUARD
          #1000 >TICKS TIMEOUT TRUE ABORT" T-Server: timeout" ENDGUARD
        ENDSELECT
        use-in CHANNEL-C@  DUP #10 U> ABORT" T-Server: TID too large"
        tids []CELL @	DUP 0= ABORT" T-Server: Unknown TID" TO ProcessorID
        ; PRIVATE

: _T800		$80000070 TO MemStart ; PRIVATE
: _T400		$80000048 TO MemStart ; PRIVATE
: _T414		$80000048 TO MemStart ; PRIVATE
: _T425		$80000048 TO MemStart ; PRIVATE
: _T200		$8024	  TO MemStart ; PRIVATE

: setparams
        ProcessorID
        CASE				\ <tid> --- <c-addr>
          #805 OF _T800 t800$ ENDOF
          #800 OF _T800 t800$ ENDOF
          #425 OF _T400 t400$ ENDOF
          #414 OF _T400 t400$ ENDOF
          #400 OF _T400 t400$ ENDOF
          #200 OF _T200 t200$ ENDOF
           TRUE ABORT" T-Server: Unknown transputer type"
        ENDCASE
        filename 0= IF COUNT TO filenamelen TO filename
                  ELSE DROP
                 ENDIF ; PRIVATE

\ SEND binary file

: sendfileblock
        LOCAL l
        0 LOCAL rl

        BEGIN	l 0>
        WHILE	l #10000 MIN TO rl
                tbuffer rl binfile READ-FILE ABORT" T-Server: Read error on file"
                        TO rl

                tbuffer rl use-out CHANNEL-SEND

                rl NEGATE +TO l
        REPEAT	; PRIVATE

: sendnulls 	0 DO 0 use-out CHANNEL-C! LOOP ; PRIVATE

: sendboot
        filename filenamelen R/O BIN OPEN-FILE	ABORT" T-Server: Unable to open file"
                TO binfile

        XYZ 5 CELLS binfile READ-FILE ABORT" T-Server: Unable to read from file"
        5 CELLS <> ABORT" T-Server: binary file too short"

        5 0 DO
                I XYZ []CELL 1+  C@ #24 LSHIFT
                I XYZ []CELL     C@ #16 LSHIFT
                I XYZ []CELL 3 + C@   8 LSHIFT
                I XYZ []CELL 2+  C@
                + + +
                I codeinfo []CELL !
          LOOP

        0 codeinfo> 5 CELLS <> ABORT" T-Server: Unrecognized header"
     {{
        3 codeinfo>  4 codeinfo> +   1 codeinfo> -  use-out CHANNEL-!
        1 codeinfo>  use-out CHANNEL-!
        MemStart use-out CHANNEL-!
        2 codeinfo>  sendfileblock
        1 codeinfo>  2 codeinfo> +  3 codeinfo> - NEGATE sendnulls
        4 codeinfo>  sendfileblock
     }}
        binfile CLOSE-FILE ABORT" T-Server: Unable to close file" ; PRIVATE


: reset	BEGIN	networktype 0=
        WHILE	." In order to reset a transputer the T-Server needs to know what type of network" CR
                ." you are using." CR
                ." Answer 1 for CSA or 2 for Parsytec: " KEY DUP EMIT CR
                CASE
                 '1' OF 1 TO networktype ENDOF
                 '2' OF 2 TO networktype ENDOF
                ENDCASE
        REPEAT
        networktype
        CASE	1 OF RESET-LINK-C ENDOF
                2 OF RESET-LINK-P ENDOF
        ENDCASE	; PRIVATE


: shouldweboot
        boot 2 <> IF EXIT ENDIF
        SELECT
          use-in GUARD TRUE  ENDGUARD
               DEFAULT FALSE ENDGUARD
        ENDSELECT
        IF 0 TO boot EXIT ENDIF
        1 TO boot ; PRIVATE

: doboot
        shouldweboot
        boot 0= IF EXIT ENDIF
        ." Booting tForth..." CR
        reset
        bootfile
        determine
        setparams
        sendboot ; PRIVATE


\ This can be done smaller and neater, but it works, so ...

VALUE f PRIVATE
CREATE filenamexyz #258 ALLOT

: +PLACE	>S cmdlptr  S CMOVE	 		\ <c-addr> <u> --- <>
                S> +TO cmdlptr ; PRIVATE

: PLACE-CR	^M cmdlptr C! 1 +TO cmdlptr ; PRIVATE 	\ <> --- <>

: +PLACE2	>S cmdincptr S CMOVE 	 		\ <c-addr> <u> --- <>
                S> +TO cmdincptr ; PRIVATE

: PLACE-CR2	^M cmdincptr C! 1 +TO cmdincptr	; PRIVATE

: +PLACE3	>S cmdextraptr S CMOVE 	 		\ <c-addr> <u> --- <>
                S> +TO cmdextraptr ; PRIVATE

: PLACE-CR3	^M cmdextraptr C!  1 +TO cmdextraptr ; PRIVATE

: cmdline
        boot 1 = IF
                memsize  0<> IF	memsize S>D <# #S #>   +PLACE
                                S"  MEMTOP ! INIT-MEM" +PLACE
                                PLACE-CR
                 	  ENDIF
                growsize 0<> IF	growsize S>D <# #S #>  +PLACE
                                S"  GROW" 	       +PLACE
                                PLACE-CR PLACE-CR
                          ENDIF

                S" tforth.prf" R/O OPEN-FILE  IF DROP 0 ENDIF TO f
                f 0<> IF  f CLOSE-FILE ABORT" problem with tforth.prf"
                          S" INCLUDE tforth.prf" +PLACE  PLACE-CR
                   ENDIF
        ENDIF

        cmdinc cmdlptr cmdincptr cmdinc - CMOVE
        cmdincptr cmdinc - +TO cmdlptr

        cmdextra cmdlptr cmdextraptr cmdextra - CMOVE
        cmdextraptr cmdextra - +TO cmdlptr ; PRIVATE


: (TFORTH)	\ <link> --- <>
        CR DEPTH 1 < ABORT" specify a LINK"
        DUP >INPUT-CHANNEL BOOTLINK-IN = ABORT" that's the boot link"
        ( link) DUP >INPUT-CHANNEL  TO use-in
                    >OUTPUT-CHANNEL TO use-out
        doboot
        boot 0= IF
                memsize  IF ." Warning: option " memsize . ." /m ignored" CR ENDIF
                growsize IF ." Warning: option " growsize . ." /g ignored" CR ENDIF
                cmdincptr cmdinc <> IF ." Warning: option(s) /i ignored" cmdinc TO cmdincptr CR ENDIF
                ."                                     tForth" CR
                CR
                ."            Copyright 1991, 1992, 1993 Dutch Transputer Forth Workshop" CR
                ." Re-connecting..." CR
        ENDIF
        cmdline
        0 TO memsize
        0 TO growsize
        cmdinc TO cmdincptr
        cmdextra TO cmdextraptr
        0 TO filenamelen 0 TO filename
        2 TO boot
        SERVE ; PRIVATE

: /b	1 TO boot ;
: /r	0 TO boot ;
: /f	TO filenamelen TO filename ;			\ <c-addr> <u> --- <>
: /g	TO growsize ;
: /i	S" INCLUDE " +PLACE2  +PLACE2  PLACE-CR2 ; 	\ <c-addr> <u> --- <>
: /c	+PLACE3  PLACE-CR3 ;				\ <c-addr> <u> --- <>
: /m	TO memsize ;

: /?	CR ." T-Server:"
        CR ." [options] <link> TFORTH <commands>  Connect over the given link"
        CR ." You can use the same options as for the shell commandline version."
        CR
        CR ." /b                  -- boot the transputer and load tForth"
        CR ." /r                  -- re-connect, do not boot "
        CR ." S'' tf8fd___.x'' /f -- use tf8fd___.x as the binary boot file"
        CR ." S'' myfile.frt'' /i -- load myfile.frt after start up"
        CR ." #1000 /g            -- expand memory available to ALLOCATE"
        CR ." $20000 /m           -- expand memory available with ALLOT"
        CR ." /h  or  /?          -- print this help" CR
        CR ." Acceptable for <link> : LINK0 LINK1 LINK2 or LINK3" ;

: /h	/? ;

: TFORTH	0 WORD C@+ DUP IF /c 		\ "/c commands"
                             ELSE 2DROP
                            ENDIF
                (TFORTH) ;

:ABOUT		/? ;

                .HELP
                DEPRIVE

                                ( * End of File * )
